==> bouncer-1.2.0a.tar.gz: bouncer-1.2.0a/bouncer.c <==
/* Original source obtained from:
 * http://www.securiteam.com/tools/3D5PTR5QAE.html
 * Also found in various places:
 * http://home.cyberarmy.com/tcu/programs/bouncer_c.txt
 *
 * $Id: bouncer.c,v 1.4 2002/09/25 07:12:05 andrewm Exp $
 *
 * Changes to original bouncer.c (c) 2002 Leading Edge Business Solutions
 * Distributable under the GNU public licence version 2 (see COPYING)
 */ 
/*
Tool Name
5/12/1999 Bouncer, a dynamic wrapper for bouncing network packets

 Details Bouncer is a tool that dynamically binds into the Linux operating
 system (other UNIX flavors might also work) allowing the administrator of that
 machine to use a third party proxy (such as WinGate, or any other SOCKS4
 supporting proxy) as the anonymous gateways to the Internet.

This can be used, for example, to fetch your mail anonymously, to transparently
bounce IRC, HTTP or even remote exploit connections. Actually, anything AF_INET
that uses SOCK_STREAM is likely to work. It does this by grabbing calls to
connect() and, behind the scenes, use one or more proxies to "bounce" and thus
hide the source of the connection (you).

The mentioned tool is attached:

*/
/*

 bouncer v1.1
 connect() client dynamic wrapper to bounce tcp connections
 by mirage (ptlink/ptnet) <mirage@hackers-pt.org>

 .\\idgard Security Services [March/April/May 1999]
 Some portability fixes added in September

 MANY thanks to LiquidK @ptnet/promiscnet/efnet
 this whole .c was based on an example of his...

 this can be used, for example, to fetch your mail anonymously, to
 transparently bounce irc, http or even remote exploit connections.
 actually, anything AF_INET that uses SOCK_STREAM is likely to work.
 it does this by grabbing calls to connect() and, behind the scenes, use one
 or more proxies to "bounce" and thus hide the source of the connection (you).

 tested only in linux, but others should work too, at least with some changes.
 it doesn't require root priviledges.

 compilation:
   linux: gcc bouncer.c -o bouncer.so -ldl -shared -O2 -s
   bsd: gcc bouncer.c -c -fPIC -O2 -s && ld bouncer.o -o bouncer.so -shared
 you may need to use -DSOCKLEN, depending on the OS and libs.

 to use, in bash:
   export BOUNCER="type:host:port@type:host:port(@...)"
   export LD_PRELOAD=/path/to/bouncer.so
   ./your_connect_program
 where <type> is:
   'w' for Wingate
   's' for SOCKS4
   'c' for CSM Proxy
   'h' for HTTP Proxy

 to unload:
   unset BOUNCER LD_PRELOAD

 some debugging output can be turned on by setting the environment variable
 BOUNCER_DEBUG to "1".

 note that, when using wingate, byte 255 is used as a telnet control caracter,
 so wingate should not be used in binary connections, not even before a socks
 bounce for example (socks may use byte 255 in its protocol). AFAIK there is
 no simple way to fix this, suggestions are welcome.

TODO: source routing proxy (says andew)

*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dlfcn.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <limits.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#ifdef SOCKLEN
typedef int socklen_t;
#endif

#define SOCKS4_USER "billgates"

#define BOUNCE_NORMAL 0
#define BOUNCE_WINGATE 1
#define BOUNCE_SOCKS4 2
#define BOUNCE_CSM 3
#define BOUNCE_HTTP 4
#define YES 1
#define NO 0
#define ERROR -1

char *bounce_env;
static void *lib_handle = NULL;
/* place to put the old connect() call */
static int (*old_connect) (int sockfd, const struct sockaddr *serv_addr, int addrlen);
/* this is not reentrant :) */
int in_use=0;
/* whether to print or not some debugging output */
int debug=0;

/* get original connect() address and BOUNCER from the environment */
static void
init (void)
{
 char *debug_env = getenv ("BOUNCER_DEBUG");

 if (debug_env != NULL)
   if (debug_env[0] == '1')
    {
     debug = YES;
     printf ("bouncer debugging started\r\n");
    }

 bounce_env = getenv ("BOUNCER");

 if (!(lib_handle = dlopen ("libc.so", RTLD_LAZY)))
   if (!(lib_handle = dlopen ("libc.so.6", RTLD_LAZY)))
    if (!(lib_handle = dlopen ("libc.so.5", RTLD_LAZY)))
     {
       printf ("error loading libc!\n");
       exit (1);
     }

 if (!(old_connect = dlsym (lib_handle, "connect")))
   {
    printf ("connect() not found in libc!\n");
    exit (1);
   }

}

/* connect to <host:port>, using the given <sock> */
int
connect_to (int sock, char *host, int port)
{
 struct hostent *he = NULL;
 struct sockaddr_in sa;

 if ((he=gethostbyname(host)) == NULL)
   return ERROR;
 memcpy (&sa.sin_addr, he->h_addr, he->h_length);
 sa.sin_port = htons (port);
 sa.sin_family = AF_INET;
 return old_connect (sock, (struct sockaddr *)&sa, sizeof sa);
}

/* read <num> chars from <sock> to <dest> */
int
read_sock (int sock, char *dest, int num)
{
 int i;
 for (i = 0; i < num; i++)
   if (read (sock, dest+i, 1) == ERROR)
    return ERROR;
 dest[i] = 0;
 return !ERROR;
}

/* read from <sock> until a <c> is received
 * returns ERROR or the number of bytes read up to and including the desired
 * character (that's new)*/
int
read_until (int sock, char c)
{
 int bytesread = 0;
 char buf = 255;
 while (buf != c) {
   if (read (sock, &buf, 1) == ERROR)
    return ERROR;
   bytesread++;
 }
 return bytesread;
}

/* bounce through a Wingate */
int
bounce_wingate (int sock, char *host, int port)
{
 char buf[80];

 if (debug)
   printf ("wingate in %s:%d\r\n", host, port);

 snprintf (buf, 80, "%s %d\n", host, port);
 write (sock, buf, strlen (buf));

 if (debug)
   printf ("request sent\r\n");

 if ( (read_until (sock, '\n') == ERROR)
     || (read_until (sock, '\n') == ERROR) )
   return ERROR;

 if (debug)
   printf ("done\r\n");

 return !ERROR;
}

/* bounce through a SOCKS4 firewall */
int
bounce_socks4 (int sock, char *host, int port)
{
 struct hostent *he = NULL;
 unsigned char buf[4];
 char buf2[100];

 if (debug)
   printf ("socks4 in %s:%d\r\n", host, port);
 if ((he = gethostbyname (host)) == NULL)
   return ERROR;
 memcpy (&buf, he->h_addr, he->h_length);

 /* taken from eggdrop's net.c:proxy_connect() */
 snprintf (buf2,sizeof(buf2)-1, "\004\001%c%c%c%c%c%c%s", (port >> 8) % 256, (port % 256),
       buf[0], buf[1], buf[2], buf[3], SOCKS4_USER);
 /* FIXME? how about ntohs instead? */
 write (sock, buf2, 8 + strlen (SOCKS4_USER) + 1);

 if (debug)
   printf ("request sent\r\n");
 if (read_sock (sock, buf2, 8) == ERROR)
   return ERROR;
 if (debug)
   printf ("done\r\n");

 return !ERROR;
}

/* bounce through a CSM Proxy */
int
bounce_csm (int sock, char *host, int port)
{
 char buf[100];

 if (debug)
   printf ("csm in %s:%d\r\n", host, port);
 if (read_until (sock, '>') == ERROR)
   return ERROR;
 snprintf (buf, sizeof(buf)-1, "%s:%d\n", host, port);
 write (sock, buf, strlen (buf));
 if (debug)
   printf ("request sent\r\n");
 if (read_until (sock, '\n') == ERROR)
   return ERROR;
 if (debug)
   printf ("done\r\n");

 return !ERROR;
}

/* bounce through a HTTP Proxy using CONNECT */
int
bounce_http (int sock, char *host, int port)
{
 char buf[100];
 int rc;

 if (debug)
   printf ("http in %s:%d\r\n", host, port);

 snprintf (buf, sizeof(buf)-1, "CONNECT %s:%d HTTP/1.1\r\n\r\n", host, port);
 write (sock, buf, strlen (buf));
 if (debug)
   printf ("request sent\r\n");

 do {
   rc = read_until(sock, '\n');
   if (rc == ERROR) return ERROR;
 } while (rc>2);  /* Read until a line of max 1 character is received -
		     CR+LF=2, or just LF=1 */
 if (debug)
   printf ("done\r\n");
 return !ERROR;
}

/* decide how to connect */
int
bounce_to (int type, int sock, char *host, int port)
{
 switch (type)
   {
    case BOUNCE_NORMAL:
     return connect_to (sock, host, port);
    case BOUNCE_WINGATE:
     return bounce_wingate (sock, host, port);
    case BOUNCE_SOCKS4:
     return bounce_socks4 (sock, host, port);
    case BOUNCE_CSM:
     return bounce_csm (sock, host, port);
    case BOUNCE_HTTP:
     return bounce_http (sock, host, port);
    default:
     return ERROR; /* should never be reached */
   }
}

/* main bouncer */
int
bounce (int sock, struct sockaddr_in *sa)
{
 int last_bounce = BOUNCE_NORMAL; /* last bounce type made */
 char *type, *host, *port, *env = strdup(bounce_env);

 /* BOUNCER variable parser */
 /* strtok allows us to make only one connection, since it destroys the connect
  * string */
 if ((type = strtok (env, ":")) != NULL)
   do
    {
     if ((host=strtok(NULL, ":")) == NULL) {
       free(env);
       return ERROR;
     }
     if ((port=strtok(NULL, ",; @")) == NULL) {
       free(env);
       return ERROR;
    }
     if (bounce_to(last_bounce, sock, host, atoi(port)) == ERROR) {
       free(env);
       return ERROR;
     }
     switch (type[0])
       {
        case 'w':
         last_bounce=BOUNCE_WINGATE;
         break;
        case 's':
         last_bounce=BOUNCE_SOCKS4;
         break;
        case 'c':
         last_bounce=BOUNCE_CSM;
         break;
        case 'h':
         last_bounce=BOUNCE_HTTP;
         break;
        default:
         return ERROR;
       }
    }
   while ((type=strtok(NULL, ":")) != NULL);
 else {
   free(env);
   return ERROR;
 }
 free(env);

 /* connect to the real destination */
 host = (char *) inet_ntoa (sa->sin_addr);
 last_bounce = bounce_to (last_bounce, sock, host, ntohs (sa->sin_port));
 in_use = NO;
 return (last_bounce);
}

/* the all mighty wrapper */
int
connect (int sockfd, const struct sockaddr *serv_addr, socklen_t addrlen)
{
 int optval = 0, optlen = sizeof (optval), tmp, flags;

 if (!lib_handle) /* If not loaded, load it now */
   init ();

 /* if it's not connecting to localhost, and socket is AF_INET */
 if ( (((struct sockaddr_in *)serv_addr)->sin_addr.s_addr != 16777343) &&
     (((struct sockaddr_in *)serv_addr)->sin_family == AF_INET) )
   /* if we're not already being used */
   if (!in_use)
    /* if BOUNCER is defined */
    if (bounce_env != NULL)
     /* if getsockopt() succeds */
     if (getsockopt (sockfd, SOL_SOCKET, SO_TYPE, &optval, &optlen) != -1)
       /* and socket is SOCK_STREAM */
       if (optval == SOCK_STREAM)
        {
         /* bounce'em baby */
         in_use = YES;
         flags = fcntl (sockfd, F_GETFL);
         if (flags & O_NONBLOCK) /* if O_NONBLOCK is set */
           fcntl (sockfd, F_SETFL, flags-O_NONBLOCK); /* unset it */
         tmp = bounce (sockfd, (struct sockaddr_in *) serv_addr);
         if (flags & O_NONBLOCK)
           fcntl (sockfd, F_SETFL, flags); /* put O_NONBLOCK back in */
         return tmp;
        }

 /* otherwise, use normal connect */
 return old_connect (sockfd, serv_addr, addrlen);
}

==> bouncer-1.2.1.tar.gz: bouncer-1.2.1/bouncer.c <==
/* Original source obtained from:
 * http://www.securiteam.com/tools/3D5PTR5QAE.html
 * Also found in various places:
 * http://home.cyberarmy.com/tcu/programs/bouncer_c.txt
 *
 * $Id: bouncer.c,v 1.6 2004/05/04 12:58:12 andrewm Exp $
 *
 * Changes to original bouncer.c (c) 2002 Leading Edge Business Solutions
 * Distributable under the GNU public licence version 2 (see COPYING)
 */ 
/*
Tool Name
5/12/1999 Bouncer, a dynamic wrapper for bouncing network packets

 Details Bouncer is a tool that dynamically binds into the Linux operating
 system (other UNIX flavors might also work) allowing the administrator of that
 machine to use a third party proxy (such as WinGate, or any other SOCKS4
 supporting proxy) as the anonymous gateways to the Internet.

This can be used, for example, to fetch your mail anonymously, to transparently
bounce IRC, HTTP or even remote exploit connections. Actually, anything AF_INET
that uses SOCK_STREAM is likely to work. It does this by grabbing calls to
connect() and, behind the scenes, use one or more proxies to "bounce" and thus
hide the source of the connection (you).

The mentioned tool is attached:

*/
/* (original source code comment)

 bouncer v1.1
 connect() client dynamic wrapper to bounce tcp connections
 by mirage (ptlink/ptnet) <mirage@hackers-pt.org>

 .\\idgard Security Services [March/April/May 1999]
 Some portability fixes added in September

 MANY thanks to LiquidK @ptnet/promiscnet/efnet
 this whole .c was based on an example of his...

 this can be used, for example, to fetch your mail anonymously, to
 transparently bounce irc, http or even remote exploit connections.
 actually, anything AF_INET that uses SOCK_STREAM is likely to work.
 it does this by grabbing calls to connect() and, behind the scenes, use one
 or more proxies to "bounce" and thus hide the source of the connection (you).

 tested only in linux, but others should work too, at least with some changes.
 it doesn't require root priviledges.

 compilation:
   linux: gcc bouncer.c -o bouncer.so -ldl -shared -O2 -s
   bsd: gcc bouncer.c -c -fPIC -O2 -s && ld bouncer.o -o bouncer.so -shared
 you may need to use -DSOCKLEN, depending on the OS and libs.

 to use, in bash:
   export BOUNCER="type:host:port@type:host:port(@...)"
   export LD_PRELOAD=/path/to/bouncer.so
   ./your_connect_program
 where <type> is:
   'w' for Wingate
   's' for SOCKS4
   'c' for CSM Proxy
   'h' for HTTP Proxy

 to unload:
   unset BOUNCER LD_PRELOAD

 some debugging output can be turned on by setting the environment variable
 BOUNCER_DEBUG to "1".

 note that, when using wingate, byte 255 is used as a telnet control caracter,
 so wingate should not be used in binary connections, not even before a socks
 bounce for example (socks may use byte 255 in its protocol). AFAIK there is
 no simple way to fix this, suggestions are welcome.

*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dlfcn.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <limits.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#ifdef SOCKLEN
typedef int socklen_t;
#endif

#define SOCKS4_USER "billgates"

#define BOUNCE_NORMAL '-'
#define BOUNCE_WINGATE 'w'
#define BOUNCE_SOCKS4 's'
#define BOUNCE_CSM 'c'
#define BOUNCE_HTTP 'h'
#define YES 1
#define NO 0
#define ERROR -1

char *bounce_env;
static void *lib_handle = NULL;
/* place to put the old connect() call */
static int (*old_connect) (int sockfd, const struct sockaddr *serv_addr, int addrlen);
/* this is not reentrant :) */
int in_use=0;
/* whether to print or not some debugging output */
int debug=0;

/* get original connect() address and BOUNCER from the environment */
static void
init (void)
{
 char *debug_env = getenv ("BOUNCER_DEBUG");

 if (debug_env != NULL)
   if (debug_env[0] == '1')
    {
     debug = YES;
     fprintf(stderr,"Bouncer debugging started\r\n");
    }

 bounce_env = getenv ("BOUNCER");

 if (!(lib_handle = dlopen ("libc.so", RTLD_LAZY)))
   if (!(lib_handle = dlopen ("libc.so.6", RTLD_LAZY)))
    if (!(lib_handle = dlopen ("libc.so.5", RTLD_LAZY)))
     {
       fprintf(stderr,"error loading libc!\n");
       exit (1);
     }

 if (!(old_connect = dlsym (lib_handle, "connect")))
   {
    fprintf(stderr,"connect() not found in libc!\n");
    exit (1);
   }

}

/* connect to <host:port>, using the given <sock> */
int
connect_to (int sock, char *host, int port)
{
 struct hostent *he = NULL;
 struct sockaddr_in sa;
 int rc;

 if ((he=gethostbyname(host)) == NULL)
   return ERROR;
 memcpy (&sa.sin_addr, he->h_addr, he->h_length);
 sa.sin_port = htons (port);
 sa.sin_family = AF_INET;
 if (debug)
   fprintf(stderr,"connect:%s:%d ...", host, port);
 rc=old_connect (sock, (struct sockaddr *)&sa, sizeof sa);
 if (debug)
   fprintf(stderr," done\r\n");
 return rc;
}

/* read <num> chars from <sock> to <dest> */
int
read_sock (int sock, char *dest, int num)
{
 int i;
 for (i = 0; i < num; i++)
   if (read (sock, dest+i, 1) == ERROR)
    return ERROR;
 dest[i] = 0;
 return !ERROR;
}

/* read from <sock> until a <c> is received
 * returns ERROR or the number of bytes read up to and including the desired
 * character (that's new)*/
int
read_until (int sock, char c)
{
 int bytesread = 0;
 char buf = 255;
 while (buf != c) {
   if (read (sock, &buf, 1) == ERROR)
    return ERROR;
   bytesread++;
 }
 return bytesread;
}

/* bounce through a Wingate */
int
bounce_wingate (int sock, char *host, int port)
{
 char buf[80];

 if (port==0) port=80;
 if (debug)
   fprintf(stderr,"wingate:%s:%d ...", host, port);

 snprintf (buf, sizeof(buf)-1, "%s %d\n", host, port);
 write (sock, buf, strlen (buf));

 if (debug)
   fprintf(stderr," request ...");

 if ( (read_until (sock, '\n') == ERROR)
     || (read_until (sock, '\n') == ERROR) )
   return ERROR;

 if (debug)
   fprintf(stderr," done\r\n");

 return !ERROR;
}

/* bounce through a SOCKS4 firewall */
int
bounce_socks4 (int sock, char *host, int port)
{
 struct hostent *he = NULL;
 unsigned char buf[4];
 char buf2[100];

 if (port==0) port=1080;
 if (debug)
   fprintf(stderr,"socks4:%s:%d ...", host, port);
 if ((he = gethostbyname (host)) == NULL)
   return ERROR;
 memcpy (&buf, he->h_addr, he->h_length);

 /* taken from eggdrop's net.c:proxy_connect() */
 snprintf (buf2,sizeof(buf2)-1, "\004\001%c%c%c%c%c%c%s", (port >> 8) % 256, (port % 256),
       buf[0], buf[1], buf[2], buf[3], SOCKS4_USER);
 /* FIXME? how about ntohs instead? */
 write (sock, buf2, 8 + strlen (SOCKS4_USER) + 1);

 if (debug)
   fprintf(stderr," request ...");
 if (read_sock (sock, buf2, 8) == ERROR)
   return ERROR;
 if (debug)
   fprintf(stderr," done\r\n");

 return !ERROR;
}

/* bounce through a CSM Proxy */
int
bounce_csm (int sock, char *host, int port)
{
 char buf[100];

 if (port==0) port=23;
 if (debug)
   fprintf(stderr,"csm:%s:%d ...", host, port);
 if (read_until (sock, '>') == ERROR)
   return ERROR;
 snprintf (buf, sizeof(buf)-1, "%s:%d\n", host, port);
 write (sock, buf, strlen (buf));
 if (debug)
   fprintf(stderr," request ...");
 if (read_until (sock, '\n') == ERROR)
   return ERROR;
 if (debug)
   fprintf(stderr," done\r\n");

 return !ERROR;
}

/* bounce through a HTTP Proxy using CONNECT */
int
bounce_http (int sock, char *host, int port)
{
 char buf[100];
 int rc;

 if (port==0) port=3128;
 if (debug)
   fprintf(stderr,"http:%s:%d ...", host, port);

 snprintf (buf, sizeof(buf)-1, "CONNECT %s:%d HTTP/1.1\r\n\r\n", host, port);
 write (sock, buf, strlen (buf));
 if (debug)
   fprintf(stderr," request ...");

 do {
   rc = read_until(sock, '\n');
   if (rc == ERROR) return ERROR;
 } while (rc>2);  /* Read until a line of max 1 character is received -
		     CR+LF=2, or just LF=1 */
 if (debug)
   fprintf(stderr," done\r\n");
 return !ERROR;
}

/* decide how to connect */
int
bounce_to (int type, int sock, char *host, int port)
{
 switch (type)
   {
    case BOUNCE_NORMAL:
     return connect_to (sock, host, port);
    case BOUNCE_WINGATE:
     return bounce_wingate (sock, host, port);
    case BOUNCE_SOCKS4:
     return bounce_socks4 (sock, host, port);
    case BOUNCE_CSM:
     return bounce_csm (sock, host, port);
    case BOUNCE_HTTP:
     return bounce_http (sock, host, port);
    default:
     return ERROR; /* should never be reached */
   }
}

/* main bouncer */
int
bounce (int sock, struct sockaddr_in *sa)
{
 int last_bounce = BOUNCE_NORMAL; /* last bounce type made */
 int this_bounce,port;
 char *p, *q, type, *host, *env = strdup(bounce_env);
 int len;

 /* BOUNCER variable parser */
 /* strtok allows us to make only one connection, since it destroys the connect
  * string */
 p = env;
 while (*p) {
	this_bounce = last_bounce;
	p+=strspn(p," @;");	/* allow any whitespace before specification */
	q=p+strcspn(p," @;");	/* find the end of the spec */
	if (q==p) break;	/* no spec */
	*q++=0;		/* null terminate the string */

	type=*p; len=strcspn(p,":"); p+=len; *p++=0;
	host=p;  len=strcspn(p,":"); p+=len; *p++=0;
	port=atoi(p);

	p=q;			/* Do the next one next time */

	if (type==0 || *host==0 ) { return ERROR; }

	switch (type) {
        case 'w':
		if (port==0) {port=80;}
		last_bounce=BOUNCE_WINGATE;
		break;
        case 's':
		if (port==0) {port=1080;}
		last_bounce=BOUNCE_SOCKS4;
		break;
        case 'c':
		if (port==0) {port=23;}
		last_bounce=BOUNCE_CSM;
		break;
        case 'h':
		if (port==0) {port=3128;}
		last_bounce=BOUNCE_HTTP;
		break;
        case 'H':
		if (port==0) {port=8080;}
		last_bounce=BOUNCE_HTTP;
		break;
        default:
		free(env);
		return ERROR;
	}
	if (bounce_to(this_bounce, sock, host, port) == ERROR) {
		if (debug) {
			fprintf(stderr,"Failed %c:%s:%d\n",this_bounce,host,port);
		}
		free(env);
		return ERROR;
	}
 }
 free(env);

 /* connect to the real destination */
 host = (char *) inet_ntoa (sa->sin_addr);
 last_bounce = bounce_to (last_bounce, sock, host, ntohs (sa->sin_port));
 in_use = NO;
 return (last_bounce);
}

/* the all mighty wrapper */
int
connect (int sockfd, const struct sockaddr *serv_addr, socklen_t addrlen)
{
 int optval = 0, optlen = sizeof (optval), tmp, flags;

 if (!lib_handle) /* If not loaded, load it now */
   init ();

 /* if it's not connecting to localhost, and socket is AF_INET */
 if ( (((struct sockaddr_in *)serv_addr)->sin_addr.s_addr != 16777343) &&
     (((struct sockaddr_in *)serv_addr)->sin_family == AF_INET) )
   /* if we're not already being used */
   if (!in_use)
    /* if BOUNCER is defined */
    if (bounce_env != NULL)
     /* if getsockopt() succeds */
     if (getsockopt (sockfd, SOL_SOCKET, SO_TYPE, &optval, &optlen) != -1)
       /* and socket is SOCK_STREAM */
       if (optval == SOCK_STREAM)
        {
         /* bounce'em baby */
         in_use = YES;
         flags = fcntl (sockfd, F_GETFL);
         if (flags & O_NONBLOCK) /* if O_NONBLOCK is set */
           fcntl (sockfd, F_SETFL, flags-O_NONBLOCK); /* unset it */
         tmp = bounce (sockfd, (struct sockaddr_in *) serv_addr);
         if (flags & O_NONBLOCK)
           fcntl (sockfd, F_SETFL, flags); /* put O_NONBLOCK back in */
         return tmp;
        }

 /* otherwise, use normal connect */
 return old_connect (sockfd, serv_addr, addrlen);
}
